import { ValidationError } from "../errors.js";
import { RawSchema } from "../schema/raw/Schema.js";
import { getBinariesFromTransaction, serializeQueryParams } from "../utils.js";
import { RQLBinary } from "./binary/RQLBinary.js";
import { RQLQueryCellRef } from "./query/references.js";
import { Transaction } from "./transaction/Transaction.js";
import {
  RawTransactionResult,
  TransactionResult,
} from "./transaction/TransactionResult.js";

/**
 * Type representing valid final values to use as parameters in a RQL query.
 *
 * For references to requests/binaries, see {@link RQLQueryCellRef} and {@link RQLQueryBinaryRef}
 *
 * @category Api
 */
export type RQLQueryParamValue = string | number | boolean | null;

/**
 * Type representing RQL parameters to send along a request.
 *
 * Each key must be referenced in the RQL query.
 *
 * @category Api
 */
export type RQLQueryParams = Record<
  string,
  RQLQueryParamValue | RQLQueryCellRef | RQLBinary
>;

/**
 * Type representing the currently logged user's data as returned by the `/current-user` route
 *
 * @category Api
 */
export interface CurrentUser {
  login: string;
  eid: number;
  dcTitle: string;
}

/**
 * Type representing the body of an error sent by CubicWeb's API.
 *
 * @category Api
 */
type ApiError = {
  title: string;
  message: string;
  data?: Record<string, unknown>;
};

/**
 * Type representing the error data sent to clients when retrieving an API error.
 *
 * It simply adds the HTTP status code in a `status` key to the regular CubicWeb api error,
 * so users can handle HTTP status codes in their code.
 *
 * @category Api
 */
export type ApiErrorResponse = {
  status: number;
} & ApiError;

/**
 * Class used to handle requests to the CubicWeb API.
 *
 * You should never have to create this object yourself.
 * Instead, retrieve this object from the methods in the {@link Client} class.
 *
 * @internal
 */
export default class Api {
  private _apiUrl: string;
  private _version = "v1";
  private _rqlApiUrl: string;
  private _schemaApiUrl: string;
  private _loginApiUrl: string;
  private _logoutApiUrl: string;
  private _currentUserApiUrl: string;
  private _binaryApiUrl: string;

  /**
   * Class used to handle requests to the CubicWeb API.
   *
   * You should never have to create this object yourself.
   * Instead, retrieve this object from the methods in the {@link Client} class.
   *
   * @param apiUrl The API url. It should not include the API version as it will be automatically added.
   * Example: if you can retrieve the schema with https://example.com/api/v1/schema,
   * use https://example.com/api as `apiUrl`.
   */
  constructor(apiUrl: string) {
    this._apiUrl = apiUrl;
    this._rqlApiUrl = this.getApiRoute("rql");
    this._schemaApiUrl = this.getApiRoute("schema");
    this._loginApiUrl = this.getApiRoute("login");
    this._logoutApiUrl = this.getApiRoute("logout");
    this._currentUserApiUrl = this.getApiRoute("current-user");
    this._binaryApiUrl = this.getApiRoute("binary");
  }

  get apiUrl() {
    return this._apiUrl;
  }

  private getApiRoute(route: string) {
    return `${this._apiUrl}/${this._version}/${route}`;
  }

  /**
   * Retrieves the CubicWeb instance's JSON schema.
   *
   * @returns The CubicWeb instance's JSON schema.
   */
  async schema(): Promise<RawSchema> {
    return nonNullFetchApi<RawSchema>(this._schemaApiUrl, {
      method: "GET",
      credentials: "include",
    });
  }

  /**
   * Tries to log in the user on the configured CubicWeb instance with the given credentials.
   *
   * @param login The user login
   * @param password The user password
   * @returns A promise returning true if the login was successful, false otherwise
   */
  async login(login: string, password: string): Promise<boolean> {
    try {
      await fetchApi<null>(this._loginApiUrl, {
        method: "POST",
        body: JSON.stringify({
          login,
          password,
        }),
        credentials: "include",
      });
      return true;
    } catch (_error) {
      return false;
    }
  }

  /**
   * Logs out the current user. Will invalidate any authentication cookie active.
   *
   * @returns A promise which will resolve when the user is properly logged out
   */
  async logout(): Promise<void> {
    await fetchApi<null>(this._logoutApiUrl, {
      method: "POST",
      credentials: "include",
    });
  }

  /**
   * Get the current user associated to this client
   *
   * @returns A promise resolving to the current user
   */
  async currentUser(): Promise<CurrentUser> {
    return await nonNullFetchApi<CurrentUser>(this._currentUserApiUrl, {
      method: "GET",
      credentials: "include",
    });
  }

  /**
   * Get the binary for the given entity attribute
   *
   * @param eid Entity eid to fetch the binary from
   * @param attribute Attribute on the entity in which the binary is stored
   * @returns A promise resolving to the binary as a Blob
   */
  async binary(eid: number, attribute: string): Promise<Blob> {
    const params = new URLSearchParams({
      eid: eid.toString(),
      attribute,
    });
    return await nonNullFetchApi<Blob>(this._binaryApiUrl + params, {
      method: "GET",
      credentials: "include",
    });
  }

  /**
   * Executes the given transaction.
   *
   * If the {@link Transaction} object contains binaries, the request is automatically
   * sent using `multipart/form-data` to allow uploading binaries.
   *
   * @param transaction Transaction to execute
   * @returns A promise returning the transaction result on success
   */
  async rql(transaction: Transaction): Promise<TransactionResult> {
    const binaries = getBinariesFromTransaction(transaction);
    try {
      if (binaries.length > 0) {
        return await this.multipartTransaction(transaction, binaries);
      } else {
        return await this.jsonTransaction(transaction);
      }
    } catch (e) {
      this.handleUserErrors(e);
    }
  }

  private getTransactionParams(transaction: Transaction) {
    return transaction.queries.map((rqlQuery) => ({
      query: rqlQuery.rql,
      params: serializeQueryParams(transaction, rqlQuery.params),
    }));
  }

  private async multipartTransaction(
    transaction: Transaction,
    binaries: RQLBinary[]
  ) {
    const jsonParams = this.getTransactionParams(transaction);
    const formData = new FormData();
    formData.append("queries", JSON.stringify(jsonParams));
    binaries.forEach((binary) => {
      formData.append(binary.uuid, binary.blob);
    });

    const resultSets = await nonNullFetchApi<RawTransactionResult>(
      this._rqlApiUrl,
      {
        method: "POST",
        credentials: "include",
        body: formData,
      }
    );
    return new TransactionResult(transaction, resultSets);
  }

  private async jsonTransaction(transaction: Transaction) {
    const resultSets = await nonNullFetchApi<RawTransactionResult>(
      this._rqlApiUrl,
      {
        method: "POST",
        credentials: "include",
        body: JSON.stringify(this.getTransactionParams(transaction)),
      }
    );
    return new TransactionResult(transaction, resultSets);
  }

  private handleUserErrors(error: unknown): never {
    if (
      typeof error === "object" &&
      error !== null &&
      "title" in error &&
      typeof error.title === "string" &&
      "message" in error &&
      typeof error.message === "string"
    ) {
      if (error.title === "ValidationError") {
        throw new ValidationError(error.message);
      }
      throw error;
    }
    throw error;
  }
}

/**
 * Behaves the same as `fetchAPI` but forbids null responses.
 * If this function detects a null response from the API,
 * the promise will be rejected with a 500 error code.
 *
 * @param input The API route to use
 * @param init Additional request parameters
 * @returns A promise with the non-null fetched result
 */
async function nonNullFetchApi<R = unknown>(
  input: RequestInfo,
  init?: RequestInit
) {
  const result = await fetchApi<R>(input, init);
  if (result == null) {
    const error: ApiErrorResponse = {
      status: 500,
      title: "Unexpected null response",
      message: "Response was null but a value was expected",
    };
    throw error;
  }
  return result;
}

/**
 * Sends a request to the given API route with the `Content-Type` header set and waits for a response.
 * The returned promise will be rejected if the API returns HTTP error codes.
 * The rejected promise will contain an `ApiErrorResponse` object.
 *
 * @param input The API route to use
 * @param init Additional request parameters
 * @returns A promise with the fetched result
 */
async function fetchApi<R = unknown>(
  input: RequestInfo,
  init?: RequestInit
): Promise<R | null> {
  let headers = init?.headers;
  if (init?.method && init.method !== "GET") {
    headers = {
      // CSRF custom header necessary for POST requests
      "X-Client-Name": "XMLHttpRequest",
    };
    // Browsers will automatically set the multipart content type
    // with the right boundary if the body is FormData
    if (!(init.body instanceof FormData)) {
      headers["Content-Type"] = "application/json;charset=UTF-8";
    }
    headers = {
      ...headers,
      ...init?.headers,
    };
  }

  const response = await fetch(input, {
    ...init,
    headers,
  });
  const status = response.status;
  if (status >= 200 && status < 300) {
    if (status === 204) {
      return null;
    }
    try {
      return await response.json();
    } catch (e) {
      console.error(e);
      return null;
    }
  } else {
    let json;
    try {
      json = await response.json();
    } catch (e) {
      console.error(e);
      throw { status, title: "Error", message: "" };
    }
    throw { status, ...json };
  }
}
