/**
 * Interface representing the json returned by the API for a single result set
 *
 * @category Api
 */
export interface RawResultSet {
  readonly rows: ResultSetCell[][];
  readonly column_names: string[];
}

/**
 * Interface representing the json returned by the API for a single row in a result set
 *
 * @category Api
 */
export type RawResultSetRow = ResultSetCell[];

/**
 * Type representing a final value in a result set.
 *
 * Please not this type may not be accurate if you use entity methods,
 * as those methods may return any data structure.
 *
 * @category Api
 */
export type ResultSetCell = string | number | boolean | null;

/**
 * Class representing a row returned in a result set.
 *
 * @category Api
 */
export class ResultSetRow implements Iterable<ResultSetCell> {
  constructor(
    public readonly columnNames: string[],
    private readonly cells: RawResultSetRow
  ) {}

  *[Symbol.iterator]() {
    for (const cell of this.cells) {
      yield cell;
    }
  }

  get length() {
    return this.cells.length;
  }

  getCell(column: number | string) {
    if (typeof column === "number") {
      return this.cells[column];
    } else {
      const columnIndex = this.columnNames.indexOf(column);
      return this.cells[columnIndex];
    }
  }
}

/**
 * Class representing a RQL request result.
 *
 * @category Api
 */
export class ResultSet {
  public readonly columnNames: string[];
  readonly #rows: RawResultSetRow[];

  constructor(rawResultSet: RawResultSet) {
    this.columnNames = rawResultSet.column_names;
    this.#rows = rawResultSet.rows;
  }

  *[Symbol.iterator]() {
    for (const r of this.#rows) {
      yield new ResultSetRow(this.columnNames, r);
    }
  }

  get length() {
    return this.#rows.length;
  }

  getRow(row: number): ResultSetRow {
    return new ResultSetRow(this.columnNames, this.#rows[row]);
  }

  getCell(row: number, column: number | string): ResultSetCell {
    if (typeof column === "number") {
      return this.#rows[row][column];
    } else {
      const columnIndex = this.columnNames.indexOf(column);
      return this.#rows[row][columnIndex];
    }
  }
}
