import { InconsistentSchemaError } from "../../errors.js";
import { EntityRawSchemaArray } from "../raw/Entities.js";
import {
  RelationDefinitionRawSchemaArray,
  Cardinality,
} from "../raw/RelationDefinitions.js";
import { IsLooseSchema, EntityRawSchemaByType } from "../raw/Utils.js";
import { EntitySchema } from "./EntitySchema.js";
import { Schema } from "./Schema.js";
import { EntityRole, isCardinalityRequired } from "./Utils.js";

/**
 * Wrapper around raw relation definition schemas from the Schema.
 * You should never have to create this object yourself.
 * Instead, retrieve these objects from the methods in the `Schema` class.
 *
 * @category Schema
 */
export class RelationDefinitionSchema<
  E extends EntityRawSchemaArray = EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E> = RelationDefinitionRawSchemaArray<E>,
  RDef extends R[number] = R[number]
> {
  /**
   * Wrapper around raw relation definition schemas from the Schema.
   * You should never have to create this object yourself.
   * Instead, retrieve these objects from the methods in the `Schema` class.
   *
   * @param schema The Schema abstractin class used to create this object
   * @param relDef The raw relation definition schema
   */
  constructor(private schema: Schema<E, R>, private relDef: RDef) {}

  /**
   * The type of the relation definition. This can also be referred to as the relation definition's name.
   */
  get type(): RDef["type"] {
    return this.relDef.type;
  }

  /**
   * A short description which can be used as documentation for the relation definition schema.
   */
  get description(): RDef["description"] {
    return this.relDef.description;
  }

  /**
   * True if the object of the relation definition schema is a primitive type.
   * In this case the relation is called an "attribute".
   */
  get final(): RDef["final"] {
    return this.relDef.final;
  }

  /**
   * The list of options set for this relation definition schema.
   * The returned type will change if this relationd definition schema
   * describes an attribute or a relation.
   */
  get options(): RDef["options"] {
    return this.relDef.options;
  }

  /**
   * The list of constraints set for this relation definition schema.
   * The returned type will change if this relationd definition schema
   * describes an attribute or a relation.
   */
  get constraints(): RDef["constraints"] {
    return this.relDef.constraints;
  }

  /**
   * The default value for this relation.
   * It can be used when creating new relations if the user did not enter a value (even if required).
   */
  get default(): RDef extends { default: infer V } ? V : undefined {
    type Default = RDef extends { default: infer V } ? V : undefined;

    return (
      "default" in this.relDef ? this.relDef.default : undefined
    ) as Default;
  }

  /**
   * The cardinality of this relation definition schema.
   */
  get cardinality(): { subject: Cardinality; object: Cardinality } {
    return {
      subject: this.relDef.cardinality[0] as Cardinality,
      object: this.relDef.cardinality[1] as Cardinality,
    };
  }

  /**
   * Gets the entity associated to this relation definition schema's subject.
   *
   * @throws InconsistentSchemaError if the relation definition schema's subject is not a valid entity
   */
  get subject(): IsLooseSchema<E> extends true
    ? EntitySchema<E, R, E[number]>
    : EntitySchema<E, R, EntityRawSchemaByType<E, RDef["subject"]>> {
    const subjectSchema = this.schema.getEntity(this.relDef.subject);
    if (subjectSchema === null) {
      throw new InconsistentSchemaError(
        `entity type ${this.relDef.subject} used as subject of relation ${this.type} is not declared`
      );
    }
    // @ts-ignore: Cannot be null as we checked for this above
    return subjectSchema;
  }

  /**
   * Gets the entity associated to this relation definition schema's object.
   *
   * @throws InconsistentSchemaError if the relation definition schema's object is not a valid entity
   */
  get object(): IsLooseSchema<E> extends true
    ? EntitySchema<E, R, E[number]>
    : EntitySchema<E, R, EntityRawSchemaByType<E, RDef["object"]>> {
    const objectSchema = this.schema.getEntity(this.relDef.object);
    if (objectSchema === null) {
      throw new InconsistentSchemaError(
        `entity type ${this.relDef.object} used as object of relation ${this.type} is not declared`
      );
    }
    // @ts-ignore: Cannot be null as we checked for this above
    return objectSchema;
  }

  /**
   * Checks if the relation is required when using the subject or object point of view.
   *
   * @param role The point of view to use
   * @returns
   */
  isRequired(role: EntityRole) {
    return isCardinalityRequired(this.cardinality[role]);
  }
}
