import { RQLQueryParams, RQLQueryParamValue } from "./api/Api.js";
import { RQLBinary } from "./api/binary/RQLBinary.js";
import { RQLQuery } from "./api/query/RQLQuery.js";
import { RQLQueryCellRef } from "./api/query/references.js";
import { Transaction } from "./api/transaction/Transaction.js";

type SerializedRQLParams = Record<
  string,
  RQLQueryParamValue | SerializedRqlQueryCellRef | SerializedRqlQueryBinaryRef
>;

interface SerializedRqlQueryCellRef {
  readonly queryIndex: number;
  readonly row: number;
  readonly column: number;
  readonly type: "query_reference";
}

interface SerializedRqlQueryBinaryRef {
  readonly ref: string;
  readonly type: "binary_reference";
}

/**
 * Retrieves the query index in the given transaction from its eid
 *
 * @param transaction The transaction to search the query in
 * @param queryUUID The UUID of the query to find
 * @returns The query index in the transaction if found
 * @throws When no query with the given UUID is found
 *
 * @internal
 */
export function getQueryIndexInTransaction(
  transaction: Transaction,
  queryUUID: string
) {
  const index = transaction.queries.findIndex((q) => q.queryUUID === queryUUID);
  if (index < 0) {
    throw new Error(
      `Cannot find query of id '${queryUUID}' in the transaction`
    );
  }
  return index;
}

/**
 * Serializes the user parameters to the format understood by the API
 *
 * @param params The params passed by the user
 * @returns A serialized version of the parameters as understood by the API
 *
 * @internal
 */
export function serializeQueryParams(
  transaction: Transaction,
  params: RQLQueryParams
) {
  const newParams: SerializedRQLParams = {};
  Object.entries(params).forEach(([k, v]) => {
    if (v instanceof RQLQueryCellRef) {
      const queryIndex = getQueryIndexInTransaction(transaction, v.queryUUID);
      newParams[k] = {
        queryIndex,
        column: v.column,
        row: v.row,
        type: "query_reference",
      } as SerializedRqlQueryCellRef;
    } else if (v instanceof RQLBinary) {
      newParams[k] = {
        ref: v.uuid,
        type: "binary_reference",
      } as SerializedRqlQueryBinaryRef;
    } else {
      newParams[k] = v;
    }
  });
  return newParams;
}

export function getBinariesFromTransaction(
  transaction: Transaction
): RQLBinary[] {
  return transaction.queries.flatMap(getBinariesFromRQLQuery);
}

export function getBinariesFromRQLQuery(query: RQLQuery): RQLBinary[] {
  return Object.values(query.params).filter(
    (p): p is RQLBinary => p instanceof RQLBinary
  );
}
