import { Client, RQLQuery, Schema, Transaction } from "@cubicweb/client";
import {
  generateCreateEntityRQL,
  generateDeleteEntitiesRQL,
  generateDeleteRelationByEidsRQL,
  generateSelectEntitiesRQL,
  generateSetAttributesByEidRQL,
  generateSetRelationByEidsRQL,
  generateUpdateEntityByEidRQL,
  resultSetToEntities,
  resultSetToEntity,
} from "@cubicweb/rql-generator";
import { useEffect, useState } from "react";

const client = new Client("http://localhost:8080/api");

// @ts-ignore Inject the client in the window to allow testing from the console
window.client = client;

function App() {
  const [schema, setSchema] = useState<Schema | undefined>();
  const [cityEid, setCityEid] = useState<number | undefined>();
  const [museumEid, setMuseumEid] = useState<number | undefined>();

  useEffect(() => {
    client.schema().then(setSchema);
  }, []);

  async function login() {
    await client.login("admin", "admin");
    console.log("Logged in");
  }

  async function clearEntities() {
    const transaction = new Transaction();
    transaction.push(new RQLQuery("DELETE Museum X", {}));
    transaction.push(new RQLQuery("DELETE City X", {}));
    await client.executeTransaction(transaction);
    console.log("Cleared entities");
  }

  async function createCity() {
    if (!schema) {
      console.warn("Schema not initialized");
      return;
    }
    const transaction = new Transaction();
    const { queries, eidRef } = generateCreateEntityRQL(schema, "City");
    transaction.push(...queries);
    const result = await client.executeTransaction(transaction);
    const newEid = result.getCell(eidRef) as number;
    setCityEid(newEid);
    console.log(`Created City#${newEid}`);
  }

  async function createMuseum() {
    if (!schema) {
      console.warn("Schema not initialized");
      return;
    }
    const transaction = new Transaction();
    const { queries, eidRef } = generateCreateEntityRQL(schema, "Museum");
    transaction.push(...queries);
    const result = await client.executeTransaction(transaction);
    const newEid = result.getCell(eidRef) as number;
    setMuseumEid(newEid);
    console.log(`Created Museum#${newEid}`);
  }

  async function createMuseumWithData() {
    if (!schema) {
      console.warn("Schema not initialized");
      return;
    }
    if (!cityEid) {
      console.log("create a city first");
      return;
    }
    const transaction = new Transaction();
    const { queries, eidRef } = generateCreateEntityRQL(schema, "Museum", {
      name: "test",
      in_city: cityEid,
    });
    transaction.push(...queries);
    const result = await client.executeTransaction(transaction);
    const newEid = result.getCell(eidRef) as number;
    setMuseumEid(newEid);
    console.log(`Created Museum#${newEid} with data`);
  }

  async function deleteMuseum() {
    if (!schema) {
      console.warn("Schema not initialized");
      return;
    }
    if (!museumEid) {
      console.warn("create a museum first");
      return;
    }
    const transaction = new Transaction();
    const queries = generateDeleteEntitiesRQL(schema, "Museum", {
      where: { eid: museumEid },
    });
    transaction.push(...queries);
    await client.executeTransaction(transaction);
    console.log(`Deleted Museum#${museumEid}`);
  }

  async function setMuseumName() {
    if (!schema) {
      console.warn("Schema not initialized");
      return;
    }
    if (!museumEid) {
      console.log("create a museum first");
      return;
    }

    const transaction = new Transaction();
    const queries = generateSetAttributesByEidRQL(schema, "Museum", museumEid, {
      name: "coucou",
    });
    transaction.push(...queries);
    await client.executeTransaction(transaction);
    console.log(`Set Museum#${museumEid} name`);
  }

  async function setMuseumInCity() {
    if (!museumEid || !cityEid) {
      console.log("create a museum and a city first");
      return;
    }

    const transaction = new Transaction();
    const queries = generateSetRelationByEidsRQL([museumEid], "in_city", [
      cityEid,
    ]);
    transaction.push(...queries);
    await client.executeTransaction(transaction);
    console.log(`Set Museum#${museumEid} in_city`);
  }

  async function updateMuseum() {
    if (!schema) {
      console.warn("Schema not initialized");
      return;
    }
    if (!museumEid || !cityEid) {
      console.log("create a museum and a city first");
      return;
    }

    const transaction = new Transaction();
    const queries = generateUpdateEntityByEidRQL(
      schema,
      "Museum",
      museumEid,
      {
        name: "updated name",
        in_city: [cityEid],
      },
      {}
    );
    transaction.push(...queries);
    await client.executeTransaction(transaction);
    console.log(`Update Museum#${museumEid}`);
  }

  async function deleteMuseumInCity() {
    if (!museumEid || !cityEid) {
      console.log("create a museum and a city first");
      return;
    }

    const transaction = new Transaction();
    const queries = generateDeleteRelationByEidsRQL([museumEid], "in_city", [
      cityEid,
    ]);
    transaction.push(...queries);
    await client.executeTransaction(transaction);
    console.log(`Deleted Museum#${museumEid} in_city City#${cityEid}`);
  }

  async function getMuseum() {
    if (!schema) {
      console.warn("Schema not initialized");
      return;
    }
    if (!museumEid) {
      console.log("create a museum first");
      return;
    }
    const transaction = new Transaction();
    const queries = generateSelectEntitiesRQL(schema, "Museum", {
      where: { eid: museumEid },
      relations: ["name", "in_city"],
    });
    transaction.push(...queries);
    const result = await client.executeTransaction(transaction);
    console.log(
      resultSetToEntity(
        schema,
        "Museum",
        ["name", "in_city"],
        result.getResultSet(queries[0])
      )
    );
  }

  async function getAllMuseum() {
    if (!schema) {
      console.warn("Schema not initialized");
      return;
    }
    const transaction = new Transaction();
    const queries = generateSelectEntitiesRQL(schema, "Museum", {
      limit: 10,
      orderBy: [{ relationName: "name", ascending: true }],
      relations: ["name", "in_city"],
    });
    transaction.push(...queries);
    const result = await client.executeTransaction(transaction);
    console.log(
      resultSetToEntities(
        schema,
        "Museum",
        ["name", "in_city"],
        result.getResultSet(queries[0])
      )
    );
  }

  return (
    <div>
      <button onClick={login}>Login</button>
      <button onClick={clearEntities}>Clear</button>
      <button onClick={createCity}>Create city</button>
      <button onClick={createMuseum}>Create museum</button>
      <button onClick={createMuseumWithData}>Create museum with data</button>
      <button onClick={getMuseum}>Get museum</button>
      <button onClick={getAllMuseum}>Get all museum</button>
      <button onClick={deleteMuseum}>Delete museum</button>
      <button onClick={setMuseumName}>Set Museum name</button>
      <button onClick={setMuseumInCity}>Set Museum in_city City</button>
      <button onClick={updateMuseum}>
        Update Museum name and in_city City
      </button>
      <button onClick={deleteMuseumInCity}>Delete Museum in_city City</button>
    </div>
  );
}

export default App;
