import {
  AttributeConstraint,
  EntitySchema,
  RelationDefinitionSchema,
} from "@cubicweb/client";
import React, { PropsWithChildren, useContext } from "react";
import { CubicWebFormData } from "../hooks/useCubicWebForm";

/**
 * Type used in {@link EntityContextProvider}
 *
 * @category Entity Context
 */
export type EntityContextType = {
  /**
   * The [entity schema](https://cubicweb.pages.logilab.fr/cubicwebjs/client/classes/EntitySchema.html)
   * from [`@cubicweb/client`](https://cubicweb.pages.logilab.fr/cubicwebjs/client/index.html)
   * */
  schema?: EntitySchema;
  /**
   * The entity values if any
   */
  values?: CubicWebFormData;
};

export const EntityContext = React.createContext<EntityContextType>({});

/**
 *
 * Context providing consumers the current
 * [entity schema](https://cubicweb.pages.logilab.fr/cubicwebjs/client/classes/EntitySchema.html)
 * and values if any.
 *
 * You will usually want to wrap your form component with this context to help creating custom inputs.
 *
 * @example
 * ```tsx
 * function Form(values: CubicWebFormData) {
 *  return (
 *    <EntityContextProvider schema={entitySchema} values={values}>
 *      <MyFormContent />
 *    <EntityContextProvider/>
 *  )
 * }
 * ```
 *
 * @param props Context values as defined in {@link EntityContextType}
 *
 * @category Entity Context
 */
export const EntityContextProvider = ({
  children,
  ...contextProps
}: PropsWithChildren<EntityContextType>) => (
  <EntityContext.Provider value={contextProps}>
    {children}
  </EntityContext.Provider>
);

/**
 * Retrieves the entity schema.
 *
 * @remarks
 * The calling component must be wrapped in an {@link EntityContextProvider}
 *
 * @returns The entity schema if any
 * @throws `Entity not initialized` error if no schema is found
 *
 * @category Entity Context
 */
export function useEntitySchema() {
  const { schema } = useContext(EntityContext);
  if (schema === undefined) {
    throw new Error("Entity not initialized");
  }
  return schema;
}

/**
 * Retreives the entity values
 *
 * @remarks
 * The calling component must be wrapped in an {@link EntityContextProvider}
 *
 * @returns The entity values if any, undefined otherwise
 *
 * @category Entity Context
 */
export function useEntityValues() {
  const { values } = useContext(EntityContext);
  return values;
}

/**
 * Retrieves the value for the given attribute
 *
 * @param name The attribute name
 *
 * @returns The attribute value if any, undefined otherwise
 *
 * @category Entity Context
 */
export function useEntityValue(name: string) {
  const { values } = useContext(EntityContext);
  return values?.[name];
}

/**
 * Retrieves the entity eid
 *
 * @returns The entity eid if any, undefined otherwise
 *
 * @category Entity Context
 */
export function useEntityEid(): number | undefined {
  return useEntityValue("eid");
}

/**
 * Retrieves attribute schema for the given attribute name
 *
 * @param name The attribute name to get the schema for
 *
 * @returns The attribute schema
 * @throws
 * `Entity not initialized` error if no schema is found.
 *
 * `Could not find attribute` error if the attribute does not exist in the schema
 *
 * @category Entity Context
 */
export function useEntityAttribute(name: string): RelationDefinitionSchema {
  const entitySchema = useEntitySchema();
  const attribute = entitySchema.attributes.find((a) => a.type === name);
  if (attribute === undefined) {
    throw new Error(
      `Could not find attribute ${name} for entity ${entitySchema.type}`
    );
  }
  return attribute;
}

/**
 * Retrieves the constraint object of the given type for the given attribute if it exists.
 *
 * @param attributeName The name of the attribute to fetch constraints for
 * @param type The type of constraint to get
 * @returns The constraint object if it exists for the given attribute, null otherwise
 * @throws
 * `Entity not initialized` error if no schema is found.
 *
 * `Could not find attribute` error if the attribute does not exist in the schema
 *
 * @category Entity Context
 */
export function useEntityAttributeConstraint<
  C extends AttributeConstraint,
  T extends C["type"]
>(
  attributeName: string,
  type: T
):
  | (C extends infer Constraint
      ? Constraint extends { type: T }
        ? Constraint
        : never
      : never)
  | null {
  const attribute = useEntityAttribute(attributeName);
  const constraint = attribute.constraints.find((c) => c.type === type);
  // @ts-expect-error can't find a way to make TS understand it works
  return constraint ?? null;
}
