import { RelationDefinitionSchema } from "@cubicweb/client";
import React, { PropsWithChildren, useContext } from "react";

export type RelationDefinitionListContextType = {
  relationDefinitionList?: Readonly<RelationDefinitionSchema>[];
  role?: "subject" | "object";
};

export const RelationDefinitionListContext =
  React.createContext<RelationDefinitionListContextType>({});

/**
 *
 * Context providing consumers the current
 * [relation definition](https://cubicweb.pages.logilab.fr/cubicwebjs/client/classes/RelationDefinitionSchema.html)
 * list.
 *
 * Components for editing relations can become quite complex spanning over several React components.
 * Use this provider as the root of your custom input to prevent prop drilling on your React components.
 *
 * @remarks
 * This context provides a list of relation definitions.
 * This is useful when your custom component must handle every object/subject combinations for a single relation type.
 *
 * @example
 * Here is an example of a context containing the relation definitions for all target entity types
 * from the point of view of the current entity.
 * 
 * ```tsx
 * function CustomRelationInput(name: string) {
 *  // Make sure to wrap your form with <EntityContextProvider/>
 *  const entitySchema = useEntitySchema()
 *  // Check if we are editing the relation from the subject or object point of view
 *  const role = getRoleFromRelationType(name)
 *  // Retrieve the list of all relations from our point of view
 *  const relations = entitySchema.matchRelationDefinitions({
        [role]: entitySchema.type,
        type: relationType,
    })
 *  return (
 *    <RelationDefinitionListContextProvider 
          relationDefinitionList={relations}
          role={role}>
 *      <MyCustomInputContent />
 *    <RelationDefinitionListContextProvider/>
 *  )
 * }
 * ```
 *
 * @param props Context values as defined in {@link EntityContextType}
 *
 * @category Relation Definition Context
 */
export const RelationDefinitionListContextProvider = (
  props: PropsWithChildren<RelationDefinitionListContextType>
) => (
  <RelationDefinitionListContext.Provider
    value={{
      relationDefinitionList: props.relationDefinitionList,
      role: props.role,
    }}
  >
    {props.children}
  </RelationDefinitionListContext.Provider>
);

/**
 * Retrieves the list of relation definition schemas
 *
 * @remarks
 * The calling component must be wrapped in a {@link RelationDefinitionListContextProvider}
 *
 * @returns The list of relation definitions
 * @throws `Relation definition list not initialized` error if no relation definition list found
 *
 * @category Relation Definition Context
 */
export function useRelationDefinitionSchemaList() {
  const { relationDefinitionList } = useContext(RelationDefinitionListContext);
  if (relationDefinitionList === undefined) {
    throw new Error("Relation definition list not initialized");
  }
  return relationDefinitionList;
}

/**
 * Retrieves the role (subject or object) of the current entity in the relation definition schema list
 *
 * @remarks
 * The calling component must be wrapped in a {@link RelationDefinitionListContextProvider}
 *
 * @returns The role (subject or object) of the current entity
 * @throws `Relation definition list not initialized` error if no role found
 *
 * @category Relation Definition Context
 */
export function useRelationDefinitionRole() {
  const { role } = useContext(RelationDefinitionListContext);
  if (role === undefined) {
    throw new Error("Relation definition role not initialized");
  }
  return role;
}
