import React from "react";
import { UseFormReturn, useForm } from "react-hook-form";
import { yupResolver } from "@hookform/resolvers/yup";
import {
  Client,
  EntityRawSchemaArray,
  EntitySchema,
  RelationDefinitionRawSchemaArray,
  Schema,
  Transaction,
  TransactionResult,
  ValidationError,
  getLooseSchema,
} from "@cubicweb/client";
import {
  CustomYupErrorsTranslations,
  getValidationSchemaFromYams,
} from "@cubicweb/yup";
import { useAvailableRelations } from "./useGetCubicWebEntity.js";
import { getCubicWebErrorData, getErrorMessages } from "../util/errors.js";
import {
  generateCreateEntityRQL,
  generateSelectEntitiesRQL,
  generateUpdateEntityByEidRQL,
  resultSetToEntity,
} from "@cubicweb/rql-generator";
// eslint-disable-next-line @typescript-eslint/no-explicit-any
export type CubicWebFormData = Record<string, any>;

/**
 * Custom errors not present in yup.
 *
 * @category General Hooks
 */
export type ErrorsTranslations = {
  /** When an unknown error occurs */
  unknownError: string;
  /** Custom yup errors */
  yup: CustomYupErrorsTranslations;
};

/**
 * Parameters to describe how to build queries to the API when handling forms.
 *
 * @category General Hooks
 */
export interface UseCubicWebFormQueryParams<
  E extends EntityRawSchemaArray,
  EntityType extends E[number]["type"]
> {
  /** The entity type you want to edit. Allows the lib to know which entity schema to return. */
  type: EntityType;
  /** The eid of the entity you want to edit. If provided, the hook will fetch the initial values. */
  eid?: number;
  /** Which attributes/relations you want to ignore while editting the entity. */
  omittedFields?: string[];
  /** True if you want to include object relations. Defaults to false. */
  includeObjectRelations?: boolean;
  /** Handler to execute when the form successfully submits */
  onSuccess?: (eid: number) => void;
}

/**
 * Return values for the useCubicWebForm hook.
 *
 * @category General Hooks
 */
export interface UseCubicWebFormData {
  /**
   * Call this function to trigger the submit operation.
   * Do not use `handleSubmit` provided by react-hook-form.
   */
  onSubmit: (e: React.FormEvent) => void;
  /** Object returned by react-hook-form [useForm](https://react-hook-form.com/docs/useform) */
  formMethods: UseFormReturn;
  /**
   * True if the form is loading, false otherwise.
   * The form may load on mount if you provided and eid to fetch the initial values,
   * and when submitting the form.
   */
  loading: boolean;
  /** The entity schema of the entity */
  entityTypeSchema: EntitySchema;
}

/**
 *
 * Provides a wrapper around `useForm` from [`react-hook-form`](https://react-hook-form.com/)
 * to use with CubicWeb.
 *
 * Along with the current entity schema and the loading state, the hook returns the `useForm` values in the
 * `formMethods` key. Use this to build your own form as you would using `react-hook-form`.
 * It also provides a custom `onSubmit` function which sends the data to the appropriate endpoints.
 * That function parses backend errors and assigns them to the correct field using
 * `react-hook-form`'s [`setError`](https://react-hook-form.com/docs/useform/seterror).
 *
 * If an eid is provided, it will fetch the entity values on load.
 *
 * You can easily specify which relations to ignore with the `omittedField` parameter.
 * By default it handles all relations/attributes, even those internal to cubicweb such as `creation_date`.
 *
 * @example
 * Here is an example usage. Please refer to the documentation of
 * [`@cubicweb/client`](https://cubicweb.pages.logilab.fr/cubicwebjs/client/index.html)
 * on how to create the schema and client parameters.
 *
 * To customize error messages, please refer to `yup.setLocale`. The `errorsTranslations` parameter is used for
 * custom messages not present in yup.
 *
 * As your instance's schema and data provider will not change in your app,
 * you will usually want to wrap this hook so you don't have to pass those parameters
 * each time.
 *
 * ```ts
 * import {instanceSchema, client, errorsTranslations} from "./my-data"
 * import {UseCubicWebFormQueryParams, useCubicWebForm} from "@cubicweb/react-form-utils"
 *
 * function useMyAppForm(queryParams: UseCubicWebFormQueryParams) {
 *     return useCubicWebForm(
 *         instanceSchema,
 *         client,
 *         errorsTranslations,
 *         queryParams
 *     )
 * }
 *
 * function MyFormComponent() {
 *     const {onSubmit, formMethods, entityTypeSchema, loading} = useMyAppForm(
 *         {
 *             type: "EntityType",
 *             eid: 1,
 *             omittedFields: ["atttribute_I_want_to_ignore"],
 *             includeObjectRelations: true,
 *             onSuccess: (newEid: Identifier) => {
 *                 console.log(`Form submit success for entity #${newEid}!`)
 *             },
 *         }
 *     )
 *     ...
 * }
 * ```
 *
 * @param schema Your instance's schema from [`@cubicweb/client`](https://cubicweb.pages.logilab.fr/cubicwebjs/client/index.html)
 * @param client Your CubicWeb client from [`@cubicweb/client`](https://cubicweb.pages.logilab.fr/cubicwebjs/client/index.html)
 * @param errorsTranslations An object for custom translations
 * @param queryParams
 * @returns
 *
 * @category General Hooks
 */
export function useCubicWebForm<
  // No need to manually set these types
  // These are automatically inferred from the parameters
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>,
  EntityType extends E[number]["type"]
>(
  schema: Schema<E, R>,
  client: Client,
  errorsTranslations: ErrorsTranslations,
  {
    type,
    eid,
    omittedFields,
    includeObjectRelations,
    onSuccess,
  }: UseCubicWebFormQueryParams<E, EntityType>
): UseCubicWebFormData {
  // It is easier to work with a loosely typed schema for generic generation
  const entityTypeSchema = getLooseSchema(schema).getEntity(type);

  if (entityTypeSchema === null) {
    throw new Error(`Could not find entity ${type}`);
  }

  let defaultAttributes: Partial<CubicWebFormData> | null = null;
  if (eid == undefined) {
    defaultAttributes = {};
    for (const attribute of entityTypeSchema.attributes) {
      if (attribute.default === null) {
        continue;
      }
      defaultAttributes = {
        [attribute.type]: attribute.default,
        ...defaultAttributes,
      };
    }
  }

  const availableRelations = useAvailableRelations(
    entityTypeSchema,
    omittedFields,
    includeObjectRelations
  );

  const [initialData, setInitialData] =
    React.useState<Partial<CubicWebFormData> | null>(defaultAttributes);

  const entityTypeParser = getValidationSchemaFromYams(
    entityTypeSchema,
    errorsTranslations.yup,
    omittedFields
  );

  const formMethods = useForm({
    defaultValues: initialData === null ? {} : initialData,
    resolver: yupResolver(entityTypeParser),
  });
  const { reset, handleSubmit, setError, clearErrors } = formMethods;
  const [loading, setLoading] = React.useState<boolean>(false);

  function setGlobalError(error: Error, message: string) {
    console.error(error); // eslint-disable-line no-console
    setError("GLOBAL", {
      type: "custom",
      message,
    });
    window.scrollTo({ top: 0, left: 0, behavior: "smooth" });
  }

  const _onSubmit = async (data: CubicWebFormData) => {
    setLoading(true);
    try {
      const updatedData = data;
      if (data.eid != undefined) {
        const transaction = new Transaction();
        const queries = generateUpdateEntityByEidRQL(
          schema,
          type,
          data.eid,
          data,
          initialData ?? {}
        );
        transaction.push(...queries);
        await client.executeTransaction(transaction);
        setInitialData(data);
      } else {
        const transaction = new Transaction();
        const { queries, eidRef } = generateCreateEntityRQL(schema, type, data);
        transaction.push(...queries);
        const result = await client.executeTransaction(transaction);
        const createdEid = result.getCell(eidRef) as number;

        updatedData.eid = createdEid;
        setInitialData(updatedData);
      }
      if (onSuccess) {
        onSuccess(updatedData.eid);
      }
    } catch (e) {
      if (e instanceof ValidationError) {
        const errorData = getCubicWebErrorData(e.message);
        // Show the error next to the corresponding field if possible
        // Fallback to global errors otherwise
        if (errorData !== undefined) {
          const errorMessages = getErrorMessages(errorData, availableRelations);

          if (errorMessages.globalErrorList.length > 0) {
            setGlobalError(
              e,
              errorMessages.globalErrorList
                .map((error) => error.message)
                .join(", ")
            );
          }

          errorMessages.localErrorList.forEach((error) => {
            setError(error.field, {
              message: error.message,
            });
          });
        }
      } else {
        setGlobalError(e as Error, errorsTranslations.unknownError);
      }
    }
    setLoading(false);
  };

  React.useEffect(() => {
    if (eid != undefined) {
      setLoading(true);
      const transaction = new Transaction();
      const queries = generateSelectEntitiesRQL(schema, type, {
        where: { eid },
        relations: availableRelations,
      });
      transaction.push(...queries);

      const resolve = (result: TransactionResult) => {
        return resultSetToEntity(
          schema,
          type,
          availableRelations,
          result.getResultSet(queries[0])
        );
      };

      client
        .executeTransaction(transaction)
        .then(resolve)
        .then((data) => {
          const strippedData = recursivelyStripEmptyValues(data) ?? {};
          reset(strippedData);
          setInitialData(strippedData);
          setLoading(false);
        });
    }
  }, [type, eid, reset, availableRelations]);

  return {
    onSubmit: (e: React.FormEvent) => {
      clearErrors();
      handleSubmit(_onSubmit)(e);
    },
    formMethods,
    loading,
    entityTypeSchema,
  };
}

// TODO extracted from data-provider. Try to find a way not to use those methods

/**
 * Recursively parses the given object and removes empty values as defined in `emptyToUndefined`.
 *
 * @param data The data to strip
 * @returns The initial data without empty values
 *
 * @category Relations Utilities
 */
export const recursivelyStripEmptyValues = <T>(data: T): T | undefined => {
  if (data == null) {
    return undefined;
  } else if (typeof data === "object" && !Array.isArray(data)) {
    const finalData: Partial<Record<string, unknown>> = {};
    Object.entries(data).forEach(([key, value]) => {
      const stripedValue = recursivelyStripEmptyValues(value);
      if (stripedValue !== undefined) {
        finalData[key] = stripedValue;
      }
    });
    return emptyToUndefined(finalData as T);
  } else if (Array.isArray(data)) {
    return emptyToUndefined(
      data
        .map(recursivelyStripEmptyValues)
        .filter((v) => v !== undefined) as unknown as T
    );
  } else {
    return emptyToUndefined(data);
  }
};

/**
 * If the given value is an empty string, array or object, returns undefined.
 * Returns the given value otherwise.
 *
 * @param v The value to convert
 * @returns The given value if not empty or undefined
 *
 * @category Relations Utilities
 */
export const emptyToUndefined = <T>(v: T) => {
  if (
    (typeof v === "string" && v === "") ||
    (Array.isArray(v) && v.length === 0) ||
    // @ts-ignore typescript did not check the first condition
    (typeof v === "object" && Object.keys(v).length === 0)
  )
    return undefined;
  else return v;
};
