import {
  EntityRawSchemaArray,
  RQLQuery,
  RelationDefinitionRawSchemaArray,
  Schema,
} from "@cubicweb/client";
import { filterToRQLQueryRestrictions } from "./utils/filter.js";
import { getEntitySchema } from "./utils/schema.js";
import { RqlQueryWhere } from "./utils/types.js";

/**
 * Generates a RQL query to delete the entities of the given type
 * for given restrictions.
 *
 * @example
 * Deleting a BlogEntry using its eid:
 *
 * ```typescript
 * generateCreateEntityRQL(
 *    SCHEMA,
 *    "BlogEntry",
 *    {
 *        where: { eid: 1209 }
 *    }
 * )
 * ```
 * produces:
 *
 * ```
 * DELETE BlogEntry X WHERE X eid %(eid)s
 * ```
 *
 * @param schema The instance schema
 * @param entityType The type of entity to delete
 * @param options Restriction to put in where clause
 * @returns An array containing a single RQL queries needed to delete entities matching the restrictions. The query of the returned array produces an empty {@link ResultSet}.
 *
 * @category Deleting entities
 */
export function generateDeleteEntitiesRQL<
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>
>(
  schema: Schema<E, R>,
  entityType: string,
  { where }: Required<RqlQueryWhere>
): RQLQuery[] {
  const entitySchema = getEntitySchema(schema, entityType);
  const filterRql = filterToRQLQueryRestrictions(where, entitySchema);
  if (filterRql.rql === "") {
    throw new Error("No valid restrictions found");
  }
  const requestStr = `DELETE ${entityType} X WHERE ${filterRql.rql}`;

  return [new RQLQuery(requestStr, filterRql.params)];
}
